# Copyright (C) 2008 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Import processor that queries the input (and doesn't import)."""

from __future__ import print_function
from typing import Dict, List, Optional, Any

from .. import (
    commands,
    processor,
)


class QueryProcessor(processor.ImportProcessor):
    """An import processor that queries the input.

    No changes to the current repository are made.
    """

    known_params = (
        commands.COMMAND_NAMES + commands.FILE_COMMAND_NAMES + [b"commit-mark"]
    )

    def __init__(
        self, params: Optional[Dict[bytes, Any]] = None, verbose: bool = False
    ) -> None:
        processor.ImportProcessor.__init__(self, params, verbose)
        self.parsed_params: Dict[bytes, Optional[List[str]]] = {}
        self.interesting_commit: Optional[str] = None
        self._finished: bool = False
        if params:
            if b"commit-mark" in params:
                commit_mark = params[b"commit-mark"]
                self.interesting_commit = (
                    commit_mark.decode("utf-8")
                    if isinstance(commit_mark, bytes)
                    else commit_mark
                )
                del params[b"commit-mark"]
            for name, value in params.items():
                if value == 1:
                    # All fields
                    fields = None
                else:
                    fields = (
                        value.split(",")
                        if isinstance(value, str)
                        else value.decode("utf-8").split(",")
                    )
                self.parsed_params[name] = fields

    def pre_handler(self, cmd: "commands.ImportCommand") -> None:
        """Hook for logic before each handler starts."""
        if self._finished:
            return
        if self.interesting_commit and cmd.name == b"commit":
            # Cast to CommitCommand to access mark and to_string
            if hasattr(cmd, "mark") and hasattr(cmd, "to_string"):
                commit_cmd = cmd
                mark_str = (
                    commit_cmd.mark.decode("utf-8")
                    if isinstance(commit_cmd.mark, bytes)
                    else str(commit_cmd.mark)
                )
                if mark_str == self.interesting_commit:
                    print(commit_cmd.to_string())
                    self._finished = True
            return
        if cmd.name in self.parsed_params:
            fields = self.parsed_params[cmd.name]
            # Convert bytes fields to str for dump_str
            str_fields = None if fields is None else fields
            str_params = {k.decode("utf-8"): v for k, v in self.parsed_params.items()}
            result_str = cmd.dump_str(str_fields, str_params, self.verbose)
            print("%s" % (result_str,))

    def progress_handler(self, cmd: "commands.ProgressCommand") -> None:
        """Process a ProgressCommand."""
        pass

    def blob_handler(self, cmd: "commands.BlobCommand") -> None:
        """Process a BlobCommand."""
        pass

    def checkpoint_handler(self, cmd: "commands.CheckpointCommand") -> None:
        """Process a CheckpointCommand."""
        pass

    def commit_handler(self, cmd: "commands.CommitCommand") -> None:
        """Process a CommitCommand."""
        pass

    def reset_handler(self, cmd: "commands.ResetCommand") -> None:
        """Process a ResetCommand."""
        pass

    def tag_handler(self, cmd: "commands.TagCommand") -> None:
        """Process a TagCommand."""
        pass

    def feature_handler(self, cmd: "commands.FeatureCommand") -> None:
        """Process a FeatureCommand."""
        feature = cmd.feature_name
        if feature not in commands.FEATURE_NAMES:
            self.warning(
                "feature %s is not supported - parsing may fail"
                % (feature.decode("utf-8"),)
            )



# Copyright (C) 2008 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Exception classes for fastimport"""

from typing import Optional

# Prefix to messages to show location information
_LOCATION_FMT = "line %(lineno)d: "


class ImportError(Exception):
    """The base exception class for all import processing exceptions."""

    _fmt: str = "Unknown Import Error"

    def __init__(self) -> None:
        super(ImportError, self).__init__(self._fmt % self.__dict__)


class ParsingError(ImportError):
    """The base exception class for all import processing exceptions."""

    _fmt = _LOCATION_FMT + "Unknown Import Parsing Error"

    def __init__(self, lineno: int) -> None:
        self.lineno = lineno
        ImportError.__init__(self)


class MissingBytes(ParsingError):
    """Raised when EOF encountered while expecting to find more bytes."""

    _fmt = (
        _LOCATION_FMT + "Unexpected EOF - expected %(expected)d bytes, found %(found)d"
    )

    def __init__(self, lineno: int, expected: int, found: int) -> None:
        self.expected = expected
        self.found = found
        ParsingError.__init__(self, lineno)


class MissingTerminator(ParsingError):
    """Raised when EOF encountered while expecting to find a terminator."""

    _fmt = _LOCATION_FMT + "Unexpected EOF - expected '%(terminator)s' terminator"

    def __init__(self, lineno: int, terminator: str) -> None:
        self.terminator = terminator
        ParsingError.__init__(self, lineno)


class InvalidCommand(ParsingError):
    """Raised when an unknown command found."""

    _fmt = _LOCATION_FMT + "Invalid command '%(cmd)s'"

    def __init__(self, lineno: int, cmd: str) -> None:
        self.cmd = cmd
        ParsingError.__init__(self, lineno)


class MissingSection(ParsingError):
    """Raised when a section is required in a command but not present."""

    _fmt = _LOCATION_FMT + "Command %(cmd)s is missing section %(section)s"

    def __init__(self, lineno: int, cmd: str, section: str) -> None:
        self.cmd = cmd
        self.section = section
        ParsingError.__init__(self, lineno)


class BadFormat(ParsingError):
    """Raised when a section is formatted incorrectly."""

    _fmt = (
        _LOCATION_FMT + "Bad format for section %(section)s in "
        "command %(cmd)s: found '%(text)s'"
    )

    def __init__(self, lineno: int, cmd: str, section: str, text: str) -> None:
        self.cmd = cmd
        self.section = section
        self.text = text
        ParsingError.__init__(self, lineno)


class InvalidTimezone(ParsingError):
    """Raised when converting a string timezone to a seconds offset."""

    _fmt = _LOCATION_FMT + "Timezone %(timezone)r could not be converted.%(reason)s"

    def __init__(
        self, lineno: int, timezone: str, reason: Optional[str] = None
    ) -> None:
        self.timezone = timezone
        if reason:
            self.reason = " " + reason
        else:
            self.reason = ""
        ParsingError.__init__(self, lineno)


class PrematureEndOfStream(ParsingError):
    """Raised when the 'done' feature was specified but missing."""

    _fmt = _LOCATION_FMT + "Stream end before 'done' command"

    def __init__(self, lineno: int) -> None:
        ParsingError.__init__(self, lineno)


class UnknownDateFormat(ImportError):
    """Raised when an unknown date format is given."""

    _fmt = "Unknown date format '%(format)s'"

    def __init__(self, format: str) -> None:
        self.format = format
        ImportError.__init__(self)


class MissingHandler(ImportError):
    """Raised when a processor can't handle a command."""

    _fmt = "Missing handler for command %(cmd)s"

    def __init__(self, cmd: str) -> None:
        self.cmd = cmd
        ImportError.__init__(self)


class UnknownParameter(ImportError):
    """Raised when an unknown parameter is passed to a processor."""

    _fmt = "Unknown parameter - '%(param)s' not in %(knowns)s"

    def __init__(self, param: str, knowns: str) -> None:
        self.param = param
        self.knowns = knowns
        ImportError.__init__(self)


class BadRepositorySize(ImportError):
    """Raised when the repository has an incorrect number of revisions."""

    _fmt = "Bad repository size - %(found)d revisions found, %(expected)d expected"

    def __init__(self, expected: int, found: int) -> None:
        self.expected = expected
        self.found = found
        ImportError.__init__(self)


class BadRestart(ImportError):
    """Raised when the import stream and id-map do not match up."""

    _fmt = (
        "Bad restart - attempted to skip commit %(commit_id)s "
        "but matching revision-id is unknown"
    )

    def __init__(self, commit_id: str) -> None:
        self.commit_id = commit_id
        ImportError.__init__(self)


class UnknownFeature(ImportError):
    """Raised when an unknown feature is given in the input stream."""

    _fmt = (
        "Unknown feature '%(feature)s' - try a later importer or an earlier data format"
    )

    def __init__(self, feature: str) -> None:
        self.feature = feature
        ImportError.__init__(self)
